/**
 *  util
 *  Author: WangCK
 *  Description:
 *  Created on 2017/7/7.
 */
import { Pool, createPool, MysqlError, escape, escapeId, PoolConnection } from 'mysql'
import { Condition, WhereStruct, FindInSetStruct, LikeStruct, WhereInStruct } from './Condition'

export type StringOrNumber = string | number

export type MysqlPoolConnection = PoolConnection

export interface MysqlConfig {
    host: string
    user: string
    port: number
    password: string
    database: string
    connectionLimit?: number
    charset?: string
    timeout?: number
}

export interface MysqlResult<T> {
    results: T
    sql: string
}

export interface ChangeResult {
    fieldCount: number
    affectedRows: number
    insertId: number
    serverStatus: number
    warningCount: number
    message: string
    protocol41: boolean
    changedRows: number
}

export type MysqlKV = {
    [key: string]: StringOrNumber
}

export const MysqlOperators = {
    Equal: "=",
    NotEqual: "!=",
    GreaterThan: ">",
    GreaterThanEqual: ">=",
    LessThan: "<",
    LessThanEqual: "<=",
}

const operators: Array<string> = [
    '=', '>', '<', '>=', '<=', '!='
]

export class Mysql {
    /**
     * MySQL连接池属性
     * @type {Pool}
     */
    private mysqlPool: Pool

    /**
     * SQL语句对象
     * @type {Condition}
     */
    private condition: Condition

    /**
     * Mysql.constructor
     * @param {MysqlConfig} config
     */
    public constructor(config: MysqlConfig) {
        /**
         * 默认开启100个连接实例
         */
        if (config.connectionLimit === undefined) {
            config.connectionLimit = 100
        }
        this.condition = new Condition()
        this.mysqlPool = createPool(config)
        this.mysqlPool.on('error', function (err) {
            throw err
        })
    }

    /**
     * 执行SQL语句
     * 可以被外部直接调用
     * @param {string} sql
     * @returns {Promise<MysqlResult<T>>}
     */
    public execute<T>(sql?: string): Promise<MysqlResult<T>> {
        return new Promise((resolve, reject): void => {
            if (sql === undefined) {
                sql = this.clear()
            }

            this.mysqlPool.query(sql, (err: MysqlError, results: any): void => {
                if (err) {
                    err.sql = sql
                    return reject(err)
                }

                return resolve({ results, sql })
            })
        })
    }

    /**
     * 开启一个事务
     * @param {Mysql} mysql
     * @returns {Promise<MysqlPoolConnection>}
     * @constructor
     */
    public static BeginTransaction(mysql: Mysql): Promise<MysqlPoolConnection> {
        return new Promise((resolve, reject) => {
            mysql.mysqlPool.getConnection((err: MysqlError, connection: MysqlPoolConnection) => {
                if (err) {
                    return reject(err)
                }

                connection.beginTransaction((err: MysqlError) => {
                    if (err) {
                        return reject(err)
                    }

                    return resolve(connection)
                })
            })
        })
    }

    /**
     * 执行一个事务Query
     * @param {MysqlPoolConnection} connection
     * @param {Mysql} query
     * @returns {Promise<MysqlResult<T>>}
     * @constructor
     */
    public static TransactionQuery<T>(connection: MysqlPoolConnection, query: Mysql): Promise<MysqlResult<T>> {
        return new Promise((resolve, reject) => {
            const sql: string = query.toSql()
            connection.query(sql, null, (err: MysqlError, results: any) => {
                if (err) {
                    err.sql = sql
                    return reject(err)
                }

                return resolve({ results, sql })
            })
        })
    }

    /**
     * 事务回滚
     * @param {MysqlPoolConnection} connection
     * @returns {Promise<any>}
     * @constructor
     */
    public static Rollback(connection: MysqlPoolConnection) {
        return new Promise((resolve, reject) => {
            connection.rollback(function (err: MysqlError) {
                if (err) {
                    return reject(err)
                }

                connection.release()
                connection = null
                return resolve()
            })
        })
    }

    /**
     * 事务提交
     * @param {MysqlPoolConnection} connection
     * @returns {Promise<any>}
     * @constructor
     */
    public static Commit(connection: MysqlPoolConnection) {
        return new Promise((resolve, reject) => {
            connection.commit((err: MysqlError) => {
                if (err) {
                    return reject(err)
                }

                connection.release()
                connection = null
                return resolve()
            })
        })
    }

    /**
     * 过滤关键参数
     * @param {string} colunm
     * @param {string} splitStr
     * @param {boolean} count
     * @returns {string}
     */
    private escapeIdSplit(colunm: string, splitStr: string, count: boolean = false) {
        const splitArr: Array<string> = colunm.split(splitStr)
        if (splitArr.length < 2) {
            if (count) {
                return "COUNT(" + escapeId(colunm) + ")"
            }

            return escapeId(colunm)
        }

        splitArr.forEach((v, k) => splitArr[k] = escapeId(v.trim()))
        if (count) {
            splitArr[0] = "COUNT(" + splitArr[0] + ")"
        }

        return splitArr.join(" " + splitStr + " ")
    }

    /**
     * 设置数据表
     * @param {string} table
     * @returns {Mysql}
     */
    public table(table: string): Mysql {
        this.condition.table = this.escapeIdSplit(table, " as ")
        return this
    }

    /**
     * 设置查询字段
     * @param {string} args
     * @returns {Mysql}
     */
    public select(...args: Array<string>): Mysql {
        args.forEach((v, k) => args[k] = this.escapeIdSplit(v, " as "))
        this.condition.select.push(...args)
        return this
    }

    /**
     * 设置关联表及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @param {"JOIN" | "LEFT JOIN" | "RIGHT JOIN"} action
     * @returns {Mysql}
     */
    public join(join: string, one: string, operator: string, two: string, action: ("JOIN" | "LEFT JOIN" | "RIGHT JOIN") = "JOIN"): Mysql {
        if (operators.indexOf(operator) < 0) {
            operator = '='
        }

        join = this.escapeIdSplit(join, ' as ')
        one = escapeId(one)
        two = escapeId(two)
        this.condition.join.push({ join, one, operator, two, action })

        return this
    }

    /**
     * 设置关联表where条件
     * @param {string} join
     * @param {string} column
     * @param {string} operator
     * @param {string | number} value
     * @returns {Mysql}
     */
    public joinOnWhere(join: string, column: string, operator: string, value: string | number): Mysql {
        if (operators.indexOf(operator) < 0) {
            operator = '='
        }

        join = this.escapeIdSplit(join, ' as ')
        column = escapeId(column)
        value = escape(value)

        if (!this.condition.joinOnWhere.has(join)) {
            this.condition.joinOnWhere.set(join, [])
        }

        this.condition.joinOnWhere.get(join).push({ column, operator, value })

        return this
    }

    /**
     * 设置左关联及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @returns {Mysql}
     */
    public leftJoin(join: string, one: string, operator: string, two: string): Mysql {
        return this.join(join, one, operator, two, "LEFT JOIN")
    }

    /**
     * 设置右关联及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @returns {Mysql}
     */
    public rightJoin(join: string, one: string, operator: string, two: string): Mysql {
        return this.join(join, one, operator, two, "RIGHT JOIN")
    }

    /**
     * 设置全局where条件
     * @param column
     * @param {StringOrNumber} operator
     * @param {StringOrNumber} value
     * @returns {Mysql}
     */
    public where(column: any, operator?: StringOrNumber, value?: StringOrNumber): Mysql {
        if (column.constructor === Object) {
            for (let i in column) {
                this.where(i, '=', column[i])
            }

            return this
        }

        if (arguments.length === 2) {
            [value, operator] = [operator, '=']
        }

        column = escapeId(column)
        this.condition.where.push({ column, operator, value })

        return this
    }

    /**
     * 设置or where条件
     * @param column
     * @param {StringOrNumber} operator
     * @param {StringOrNumber} value
     * @returns {Mysql}
     */
    public orWhere(column: any, operator?: StringOrNumber, value?: StringOrNumber): Mysql {
        if (column.constructor === Object) {
            for (let i in column) {
                this.orWhere(i, '=', column[i])
            }

            return this
        }

        if (arguments.length === 2) {
            [value, operator] = [operator, '=']
        }

        column = escapeId(column)
        this.condition.orWhere.push({ column, operator, value })

        return this
    }

    /**
     * 设置where null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    public whereNull(colunm: string): Mysql {
        this.condition.whereNull.push(escapeId(colunm))
        return this
    }

    /**
     * 设置where not null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    public whereNotNull(colunm: string): Mysql {
        this.condition.whereNotNull.push(escapeId(colunm))
        return this
    }

    /**
     * 设置or where null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    public orWhereNull(colunm: string): Mysql {
        this.condition.orWhereNull.push(escapeId(colunm))
        return this
    }

    /**
     * 设置or where not null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    public orWhereNotNull(colunm: string): Mysql {
        this.condition.orWhereNotNull.push(escapeId(colunm))
        return this
    }

    /**
     * 设置order by字段
     * @param {string} column
     * @param {"DESC" | "ASC"} direction
     * @returns {Mysql}
     */
    public orderBy(column: string, direction: ('DESC' | 'ASC') = 'ASC'): Mysql {
        this.condition.order.push(this.escapeIdSplit(column, ",") + ' ' + direction)
        return this
    }

    /**
     * 设置跳过行数
     * @param {number} value
     * @returns {Mysql}
     */
    public skip(value: number): Mysql {
        this.condition.skip = value
        return this
    }

    /**
     * 设置获取行数
     * @param {number} value
     * @returns {Mysql}
     */
    public take(value: number): Mysql {
        this.condition.take = value
        return this
    }

    /**
     * 快捷设置分页参数
     * @param {number} skip
     * @param {number} take
     * @returns {Mysql}
     */
    public limit(skip: number, take?: number): Mysql {
        this.condition.skip = skip
        if (take !== undefined) {
            this.condition.take = take
        }

        return this
    }

    /**
     * 设置GroupBy分组条件
     * @param {string} value
     * @returns {Mysql}
     */
    public groupBy(value: string): Mysql {
        this.condition.groupBy = escapeId(value)
        return this
    }

    /**
     * 查询多条数据
     * @returns {Mysql}
     */
    public get(): Mysql {
        let sql: string = this.toSelectSql()
        this.condition.sql = sql
        return this
    }

    /**
     * 查询总数（聚合）
     * @param {string} field
     * @returns {Mysql}
     */
    public count(field: string): Mysql {
        this.condition.select = [this.escapeIdSplit(field, " as ", true)]
        this.get()
        return this
    }

    /**
     * 查询单条数据
     * @returns {Mysql}
     */
    public first(): Mysql {
        this.skip(1).get()
        return this
    }

    /**
     * 将Query转换为SQL
     * @returns {string}
     */
    public toSql(): string {
        return this.clear()
    }

    /**
     * 将select转换为SQL
     * @returns {string}
     */
    private toSelectSql(): string {
        let sql: string = ''

        if (this.condition.select.length < 1) {
            this.select('*')
        }
        sql += 'SELECT ' + this.condition.select.join(', ')
        sql += ' FROM ' + this.condition.table

        sql += this.toJoin()
        sql += this.toWhere()
        sql += this.toOrder()

        if (this.condition.groupBy != undefined) {
            sql += ' GROUP BY ' + this.condition.groupBy
        }

        if (this.condition.skip != undefined) {
            sql += ' LIMIT ' + this.condition.skip

            if (this.condition.take != undefined) {
                sql += ', ' + this.condition.take
            }
        }

        return sql
    }

    /**
     * 多where条件组装
     * @param {Array<T>} whereMap
     * @param {(chunk: T) => string} handle
     * @returns {string}
     */
    private whereHandle<T>(whereMap: Array<T>, handle: (chunk: T) => string): string {
        let whereList: Array<string> = []
        whereList = whereMap.map((chunk: T) => handle(chunk))
        return whereList.join(' AND ')
    }

    /**
     * 组装where条件
     * @param {string} where
     * @param {string} condition
     * @returns {string}
     */
    private joinWhere(where: string, condition: string): string {
        where += (where === '' ? ' WHERE ' : ' AND ') + condition
        return where
    }

    /**
     * 组装orWhere条件
     * @param {string} where
     * @param {string} condition
     * @returns {string}
     */
    private joinOrWhere(where: string, condition: string): string {
        where += (where === '' ? ' OR ' : ' AND ') + condition
        return where
    }

    /**
     * 将where条件转换为SQL
     * @returns {string}
     */
    private toWhere(): string {
        let where: string = ''
        let orWhere: string = ''

        if (this.condition.where.length > 0) {
            where = this.joinWhere(where, this.whereHandle<WhereStruct>(this.condition.where, chunk => {
                return chunk.column + ' ' + chunk.operator + ' ' + escape(chunk.value)
            }))
        }

        if (this.condition.whereNull.length > 0) {
            where = this.joinWhere(where, this.whereHandle<string>(this.condition.whereNull, chunk => {
                return chunk + ' IS NULL'
            }))
        }

        if (this.condition.whereNotNull.length > 0) {
            where = this.joinWhere(where, this.whereHandle<string>(this.condition.whereNotNull, chunk => {
                return chunk + ' IS NOT NULL'
            }))
        }

        if (this.condition.orWhere.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle<WhereStruct>(this.condition.orWhere, chunk => {
                return chunk.column + ' ' + chunk.operator + ' ' + escape(chunk.value)
            }))
        }

        if (this.condition.orWhereNull.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle<string>(this.condition.orWhereNull, chunk => {
                return chunk + ' IS NULL'
            }))
        }

        if (this.condition.orWhereNotNull.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle<string>(this.condition.orWhereNotNull, chunk => {
                return chunk + ' IS NOT NULL'
            }))
        }

        if (this.condition.findInSet.length > 0) {
            where = this.joinWhere(where, this.whereHandle<FindInSetStruct>(this.condition.findInSet, chunk => {
                return 'FIND_IN_SET(' + chunk.findStr + ',' + chunk.strList + ')'
            }))
        }

        if (this.condition.orFindInSet.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle<FindInSetStruct>(this.condition.orFindInSet, chunk => {
                return 'FIND_IN_SET(' + chunk.findStr + ',' + chunk.strList + ')'
            }))
        }

        if (this.condition.like.length > 0) {
            where = this.joinWhere(where, this.whereHandle<LikeStruct>(this.condition.like, chunk => {
                return chunk.field + ' LIKE ' + chunk.likeStr
            }))
        }

        if (this.condition.orLike.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle<LikeStruct>(this.condition.orLike, chunk => {
                return chunk.field + ' LIKE ' + chunk.likeStr
            }))
        }

        if (this.condition.whereIn.length > 0) {
            where = this.joinWhere(where, this.whereHandle<WhereInStruct>(this.condition.whereIn, chunk => {
                return chunk.field + ' IN (' + chunk.whereIn.join(',') + ')'
            }))
        }

        if (this.condition.orWhereIn.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle<WhereInStruct>(this.condition.orWhereIn, chunk => {
                return chunk.field + ' IN (' + chunk.whereIn.join(',') + ')'
            }))
        }

        return where + orWhere
    }

    /**
     * 将OrderBy条件转换为SQL
     * @returns {string}
     */
    private toOrder(): string {
        let order: string = ''

        if (this.condition.order.length > 0) {
            order = ' ORDER BY ' + this.condition.order.join(', ')
        }

        return order
    }

    /**
     * 将join转换为SQL
     * @returns {string}
     */
    private toJoin(): string {
        let join: string = ''

        if (this.condition.join.length > 0) {
            let joinList: Array<string> = []

            this.condition.join.forEach(chunk => {
                let joinListStr: string = ' ' + chunk.action + ' ' + chunk.join + ' ON ' + chunk.one + ' ' + chunk.operator + ' ' + chunk.two

                if (this.condition.joinOnWhere.has(chunk.join)) {
                    let joinOnWhere = this.condition.joinOnWhere.get(chunk.join)

                    if (joinOnWhere.length > 0) {
                        let joinOnWhereList: Array<string> = []
                        joinOnWhere.forEach(value => joinOnWhereList.push(value.column + ' ' + value.operator + ' ' + value.value))
                        joinListStr += ' AND ' + joinOnWhereList.join(' AND ')
                    }
                }

                joinList.push(joinListStr)
            })

            join += joinList.join(' ')
        }

        return join
    }

    /**
     * 更新数据方法
     * @param {MysqlKV} column
     * @returns {Mysql}
     */
    public update(column: MysqlKV): Mysql {
        if (column.constructor !== Object) {
            throw new Error('parameter must be a Object')
        }

        /**
         * 使用对象拷贝
         * @type {MysqlKV}
         */
        const column_copy = Object.assign({}, column)
        let sql: string = 'UPDATE ' + this.condition.table

        let update: Array<string> = []
        for (let i in column_copy) {
            update.push(escapeId(i) + ' = ' + escape(column_copy[i]))
        }

        sql += ' SET ' + update.join(', ')

        sql += this.toWhere()
        this.condition.sql = sql

        return this
    }

    /**
     * 递增方法，可设置递增数量，默认为1
     * @param {string} field
     * @param {number} incrementNum
     * @returns {Mysql}
     */
    public increment(field: string, incrementNum: number = 1): Mysql {
        let sql: string = ''
        sql += 'UPDATE ' + this.condition.table

        field = escapeId(field)
        sql += ' SET ' + field + ' = ' + field + " + " + escape(incrementNum) + " "
        sql += this.toWhere()
        this.condition.sql = sql

        return this
    }

    /**
     * 递减方法，可设置递减数量，默认为1
     * @param {string} field
     * @param {number} decrementNum
     * @returns {Mysql}
     */
    public decrement(field: string, decrementNum: number = 1): Mysql {
        let sql: string = ''
        sql += 'UPDATE ' + this.condition.table

        field = escapeId(field)
        sql += ' SET ' + field + ' = ' + field + " - " + escape(decrementNum) + " "
        sql += this.toWhere()
        this.condition.sql = sql

        return this
    }

    /**
     * SQL中findInSet方法封装
     * @param {string | number} findStr
     * @param {string} strList
     * @returns {Mysql}
     */
    public findInSet(findStr: string | number, strList: string): Mysql {
        this.condition.findInSet.push({
            findStr: escape(findStr),
            strList: escapeId(strList)
        })

        return this
    }

    /**
     * SQL中or findInSet方法封装
     * @param {string | number} findStr
     * @param {string} strList
     * @returns {Mysql}
     */
    public orFindInSet(findStr: string | number, strList: string): Mysql {
        this.condition.orFindInSet.push({
            findStr: escape(findStr),
            strList: escapeId(strList)
        })

        return this
    }

    /**
     * like字符串匹配
     * @param {string} field
     * @param {string} likeStr
     * @returns {Mysql}
     */
    public like(field: string, likeStr: string): Mysql {
        this.condition.like.push({
            field: escapeId(field),
            likeStr: escape(likeStr)
        })

        return this
    }

    /**
     * or like字符串匹配
     * @param {string} field
     * @param {string} likeStr
     * @returns {Mysql}
     */
    public orLike(field: string, likeStr: string): Mysql {
        this.condition.orLike.push({
            field: escapeId(field),
            likeStr: escape(likeStr)
        })

        return this
    }

    /**
     * where in查询方法
     * @param {string} field
     * @param {Array<string | number>} whereIn
     * @returns {Mysql}
     */
    public whereIn(field: string, whereIn: Array<string | number>): Mysql {
        /**
         * 使用数组拷贝
         * @type {string | number[]}
         */
        const whereIn_copy: Array<string | number> = [...whereIn]
        whereIn_copy.forEach((value, key) => whereIn_copy[key] = escape(value))

        this.condition.whereIn.push({
            field: escapeId(field),
            whereIn: whereIn_copy
        })

        return this
    }

    /**
     * or where in查询方法
     * @param {string} field
     * @param {Array<string | number>} whereIn
     * @returns {Mysql}
     */
    public orWhereIn(field: string, whereIn: Array<string | number>): Mysql {
        /**
         * 使用数组拷贝
         * @type {string | number[]}
         */
        const whereIn_copy: Array<string | number> = [...whereIn]
        whereIn_copy.forEach((value, key) => whereIn[key] = escape(value))

        this.condition.orWhereIn.push({
            field: escapeId(field),
            whereIn: whereIn_copy
        })

        return this
    }

    /**
     * 删除数据方法
     * @returns {Mysql}
     */
    public delete(): Mysql {
        let sql: string = ''

        sql += 'DELETE FROM ' + this.condition.table
        sql += this.toWhere()

        this.condition.sql = sql

        return this
    }

    /**
     * 插入数据方法
     * @param {MysqlKV} column
     * @returns {Mysql}
     */
    public insert(column: MysqlKV): Mysql {
        if (column.constructor !== Object) {
            throw new Error('parameter must be a Object')
        }

        /**
         * 使用对象拷贝
         * @type {MysqlKV}
         */
        const column_copy = Object.assign({}, column)
        let key: Array<string> = []
        let value: Array<StringOrNumber> = []

        for (let i in column_copy) {
            key.push(escapeId(i))
            value.push(escape(column_copy[i]))
        }

        let sql: string = 'INSERT INTO ' + this.condition.table
        sql += '(' + key.join(', ') + ')'
        sql += ' VALUES(' + value.join(', ') + ')'

        this.condition.sql = sql

        return this
    }

    /**
     * 清空Query对象方法
     * @returns {string}
     */
    public clear(): string {
        let sql: string = this.condition.sql
        this.condition.clear()
        return sql
    }
}
