/**
 *  util
 *  Author: WangCK
 *  Description:
 *  Created on 2017/7/7.
 */
import { PoolConnection } from 'mysql';
export declare type StringOrNumber = string | number;
export declare type MysqlPoolConnection = PoolConnection;
export interface MysqlConfig {
    host: string;
    user: string;
    port: number;
    password: string;
    database: string;
    connectionLimit?: number;
    charset?: string;
    timeout?: number;
}
export interface MysqlResult<T> {
    results: T;
    sql: string;
}
export interface ChangeResult {
    fieldCount: number;
    affectedRows: number;
    insertId: number;
    serverStatus: number;
    warningCount: number;
    message: string;
    protocol41: boolean;
    changedRows: number;
}
export declare type MysqlKV = {
    [key: string]: StringOrNumber;
};
export declare const MysqlOperators: {
    Equal: string;
    NotEqual: string;
    GreaterThan: string;
    GreaterThanEqual: string;
    LessThan: string;
    LessThanEqual: string;
};
export declare class Mysql {
    /**
     * MySQL连接池属性
     * @type {Pool}
     */
    private mysqlPool;
    /**
     * SQL语句对象
     * @type {Condition}
     */
    private condition;
    /**
     * Mysql.constructor
     * @param {MysqlConfig} config
     */
    constructor(config: MysqlConfig);
    /**
     * 执行SQL语句
     * 可以被外部直接调用
     * @param {string} sql
     * @returns {Promise<MysqlResult<T>>}
     */
    execute<T>(sql?: string): Promise<MysqlResult<T>>;
    /**
     * 开启一个事务
     * @param {Mysql} mysql
     * @returns {Promise<MysqlPoolConnection>}
     * @constructor
     */
    static BeginTransaction(mysql: Mysql): Promise<MysqlPoolConnection>;
    /**
     * 执行一个事务Query
     * @param {MysqlPoolConnection} connection
     * @param {Mysql} query
     * @returns {Promise<MysqlResult<T>>}
     * @constructor
     */
    static TransactionQuery<T>(connection: MysqlPoolConnection, query: Mysql): Promise<MysqlResult<T>>;
    /**
     * 事务回滚
     * @param {MysqlPoolConnection} connection
     * @returns {Promise<any>}
     * @constructor
     */
    static Rollback(connection: MysqlPoolConnection): Promise<{}>;
    /**
     * 事务提交
     * @param {MysqlPoolConnection} connection
     * @returns {Promise<any>}
     * @constructor
     */
    static Commit(connection: MysqlPoolConnection): Promise<{}>;
    /**
     * 过滤关键参数
     * @param {string} colunm
     * @param {string} splitStr
     * @param {boolean} count
     * @returns {string}
     */
    private escapeIdSplit;
    /**
     * 设置数据表
     * @param {string} table
     * @returns {Mysql}
     */
    table(table: string): Mysql;
    /**
     * 设置查询字段
     * @param {string} args
     * @returns {Mysql}
     */
    select(...args: Array<string>): Mysql;
    /**
     * 设置关联表及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @param {"JOIN" | "LEFT JOIN" | "RIGHT JOIN"} action
     * @returns {Mysql}
     */
    join(join: string, one: string, operator: string, two: string, action?: ("JOIN" | "LEFT JOIN" | "RIGHT JOIN")): Mysql;
    /**
     * 设置关联表where条件
     * @param {string} join
     * @param {string} column
     * @param {string} operator
     * @param {string | number} value
     * @returns {Mysql}
     */
    joinOnWhere(join: string, column: string, operator: string, value: string | number): Mysql;
    /**
     * 设置左关联及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @returns {Mysql}
     */
    leftJoin(join: string, one: string, operator: string, two: string): Mysql;
    /**
     * 设置右关联及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @returns {Mysql}
     */
    rightJoin(join: string, one: string, operator: string, two: string): Mysql;
    /**
     * 设置全局where条件
     * @param column
     * @param {StringOrNumber} operator
     * @param {StringOrNumber} value
     * @returns {Mysql}
     */
    where(column: any, operator?: StringOrNumber, value?: StringOrNumber): Mysql;
    /**
     * 设置or where条件
     * @param column
     * @param {StringOrNumber} operator
     * @param {StringOrNumber} value
     * @returns {Mysql}
     */
    orWhere(column: any, operator?: StringOrNumber, value?: StringOrNumber): Mysql;
    /**
     * 设置where null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    whereNull(colunm: string): Mysql;
    /**
     * 设置where not null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    whereNotNull(colunm: string): Mysql;
    /**
     * 设置or where null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    orWhereNull(colunm: string): Mysql;
    /**
     * 设置or where not null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    orWhereNotNull(colunm: string): Mysql;
    /**
     * 设置order by字段
     * @param {string} column
     * @param {"DESC" | "ASC"} direction
     * @returns {Mysql}
     */
    orderBy(column: string, direction?: ('DESC' | 'ASC')): Mysql;
    /**
     * 设置跳过行数
     * @param {number} value
     * @returns {Mysql}
     */
    skip(value: number): Mysql;
    /**
     * 设置获取行数
     * @param {number} value
     * @returns {Mysql}
     */
    take(value: number): Mysql;
    /**
     * 快捷设置分页参数
     * @param {number} skip
     * @param {number} take
     * @returns {Mysql}
     */
    limit(skip: number, take?: number): Mysql;
    /**
     * 设置GroupBy分组条件
     * @param {string} value
     * @returns {Mysql}
     */
    groupBy(value: string): Mysql;
    /**
     * 查询多条数据
     * @returns {Mysql}
     */
    get(): Mysql;
    /**
     * 查询总数（聚合）
     * @param {string} field
     * @returns {Mysql}
     */
    count(field: string): Mysql;
    /**
     * 查询单条数据
     * @returns {Mysql}
     */
    first(): Mysql;
    /**
     * 将Query转换为SQL
     * @returns {string}
     */
    toSql(): string;
    /**
     * 将select转换为SQL
     * @returns {string}
     */
    private toSelectSql;
    /**
     * 多where条件组装
     * @param {Array<T>} whereMap
     * @param {(chunk: T) => string} handle
     * @returns {string}
     */
    private whereHandle;
    /**
     * 组装where条件
     * @param {string} where
     * @param {string} condition
     * @returns {string}
     */
    private joinWhere;
    /**
     * 组装orWhere条件
     * @param {string} where
     * @param {string} condition
     * @returns {string}
     */
    private joinOrWhere;
    /**
     * 将where条件转换为SQL
     * @returns {string}
     */
    private toWhere;
    /**
     * 将OrderBy条件转换为SQL
     * @returns {string}
     */
    private toOrder;
    /**
     * 将join转换为SQL
     * @returns {string}
     */
    private toJoin;
    /**
     * 更新数据方法
     * @param {MysqlKV} column
     * @returns {Mysql}
     */
    update(column: MysqlKV): Mysql;
    /**
     * 递增方法，可设置递增数量，默认为1
     * @param {string} field
     * @param {number} incrementNum
     * @returns {Mysql}
     */
    increment(field: string, incrementNum?: number): Mysql;
    /**
     * 递减方法，可设置递减数量，默认为1
     * @param {string} field
     * @param {number} decrementNum
     * @returns {Mysql}
     */
    decrement(field: string, decrementNum?: number): Mysql;
    /**
     * SQL中findInSet方法封装
     * @param {string | number} findStr
     * @param {string} strList
     * @returns {Mysql}
     */
    findInSet(findStr: string | number, strList: string): Mysql;
    /**
     * SQL中or findInSet方法封装
     * @param {string | number} findStr
     * @param {string} strList
     * @returns {Mysql}
     */
    orFindInSet(findStr: string | number, strList: string): Mysql;
    /**
     * like字符串匹配
     * @param {string} field
     * @param {string} likeStr
     * @returns {Mysql}
     */
    like(field: string, likeStr: string): Mysql;
    /**
     * or like字符串匹配
     * @param {string} field
     * @param {string} likeStr
     * @returns {Mysql}
     */
    orLike(field: string, likeStr: string): Mysql;
    /**
     * where in查询方法
     * @param {string} field
     * @param {Array<string | number>} whereIn
     * @returns {Mysql}
     */
    whereIn(field: string, whereIn: Array<string | number>): Mysql;
    /**
     * or where in查询方法
     * @param {string} field
     * @param {Array<string | number>} whereIn
     * @returns {Mysql}
     */
    orWhereIn(field: string, whereIn: Array<string | number>): Mysql;
    /**
     * 删除数据方法
     * @returns {Mysql}
     */
    delete(): Mysql;
    /**
     * 插入数据方法
     * @param {MysqlKV} column
     * @returns {Mysql}
     */
    insert(column: MysqlKV): Mysql;
    /**
     * 清空Query对象方法
     * @returns {string}
     */
    clear(): string;
}
