"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 *  util
 *  Author: WangCK
 *  Description:
 *  Created on 2017/7/7.
 */
const mysql_1 = require("mysql");
const Condition_1 = require("./Condition");
exports.MysqlOperators = {
    Equal: "=",
    NotEqual: "!=",
    GreaterThan: ">",
    GreaterThanEqual: ">=",
    LessThan: "<",
    LessThanEqual: "<=",
};
const operators = [
    '=', '>', '<', '>=', '<=', '!='
];
class Mysql {
    /**
     * Mysql.constructor
     * @param {MysqlConfig} config
     */
    constructor(config) {
        /**
         * 默认开启100个连接实例
         */
        if (config.connectionLimit === undefined) {
            config.connectionLimit = 100;
        }
        this.condition = new Condition_1.Condition();
        this.mysqlPool = mysql_1.createPool(config);
        this.mysqlPool.on('error', function (err) {
            throw err;
        });
    }
    /**
     * 执行SQL语句
     * 可以被外部直接调用
     * @param {string} sql
     * @returns {Promise<MysqlResult<T>>}
     */
    execute(sql) {
        return new Promise((resolve, reject) => {
            if (sql === undefined) {
                sql = this.clear();
            }
            this.mysqlPool.query(sql, (err, results) => {
                if (err) {
                    err.sql = sql;
                    return reject(err);
                }
                return resolve({ results, sql });
            });
        });
    }
    /**
     * 开启一个事务
     * @param {Mysql} mysql
     * @returns {Promise<MysqlPoolConnection>}
     * @constructor
     */
    static BeginTransaction(mysql) {
        return new Promise((resolve, reject) => {
            mysql.mysqlPool.getConnection((err, connection) => {
                if (err) {
                    return reject(err);
                }
                connection.beginTransaction((err) => {
                    if (err) {
                        return reject(err);
                    }
                    return resolve(connection);
                });
            });
        });
    }
    /**
     * 执行一个事务Query
     * @param {MysqlPoolConnection} connection
     * @param {Mysql} query
     * @returns {Promise<MysqlResult<T>>}
     * @constructor
     */
    static TransactionQuery(connection, query) {
        return new Promise((resolve, reject) => {
            const sql = query.toSql();
            connection.query(sql, null, (err, results) => {
                if (err) {
                    err.sql = sql;
                    return reject(err);
                }
                return resolve({ results, sql });
            });
        });
    }
    /**
     * 事务回滚
     * @param {MysqlPoolConnection} connection
     * @returns {Promise<any>}
     * @constructor
     */
    static Rollback(connection) {
        return new Promise((resolve, reject) => {
            connection.rollback(function (err) {
                if (err) {
                    return reject(err);
                }
                connection.release();
                connection = null;
                return resolve();
            });
        });
    }
    /**
     * 事务提交
     * @param {MysqlPoolConnection} connection
     * @returns {Promise<any>}
     * @constructor
     */
    static Commit(connection) {
        return new Promise((resolve, reject) => {
            connection.commit((err) => {
                if (err) {
                    return reject(err);
                }
                connection.release();
                connection = null;
                return resolve();
            });
        });
    }
    /**
     * 过滤关键参数
     * @param {string} colunm
     * @param {string} splitStr
     * @param {boolean} count
     * @returns {string}
     */
    escapeIdSplit(colunm, splitStr, count = false) {
        const splitArr = colunm.split(splitStr);
        if (splitArr.length < 2) {
            if (count) {
                return "COUNT(" + mysql_1.escapeId(colunm) + ")";
            }
            return mysql_1.escapeId(colunm);
        }
        splitArr.forEach((v, k) => splitArr[k] = mysql_1.escapeId(v.trim()));
        if (count) {
            splitArr[0] = "COUNT(" + splitArr[0] + ")";
        }
        return splitArr.join(" " + splitStr + " ");
    }
    /**
     * 设置数据表
     * @param {string} table
     * @returns {Mysql}
     */
    table(table) {
        this.condition.table = this.escapeIdSplit(table, " as ");
        return this;
    }
    /**
     * 设置查询字段
     * @param {string} args
     * @returns {Mysql}
     */
    select(...args) {
        args.forEach((v, k) => args[k] = this.escapeIdSplit(v, " as "));
        this.condition.select.push(...args);
        return this;
    }
    /**
     * 设置关联表及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @param {"JOIN" | "LEFT JOIN" | "RIGHT JOIN"} action
     * @returns {Mysql}
     */
    join(join, one, operator, two, action = "JOIN") {
        if (operators.indexOf(operator) < 0) {
            operator = '=';
        }
        join = this.escapeIdSplit(join, ' as ');
        one = mysql_1.escapeId(one);
        two = mysql_1.escapeId(two);
        this.condition.join.push({ join, one, operator, two, action });
        return this;
    }
    /**
     * 设置关联表where条件
     * @param {string} join
     * @param {string} column
     * @param {string} operator
     * @param {string | number} value
     * @returns {Mysql}
     */
    joinOnWhere(join, column, operator, value) {
        if (operators.indexOf(operator) < 0) {
            operator = '=';
        }
        join = this.escapeIdSplit(join, ' as ');
        column = mysql_1.escapeId(column);
        value = mysql_1.escape(value);
        if (!this.condition.joinOnWhere.has(join)) {
            this.condition.joinOnWhere.set(join, []);
        }
        this.condition.joinOnWhere.get(join).push({ column, operator, value });
        return this;
    }
    /**
     * 设置左关联及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @returns {Mysql}
     */
    leftJoin(join, one, operator, two) {
        return this.join(join, one, operator, two, "LEFT JOIN");
    }
    /**
     * 设置右关联及字段
     * @param {string} join
     * @param {string} one
     * @param {string} operator
     * @param {string} two
     * @returns {Mysql}
     */
    rightJoin(join, one, operator, two) {
        return this.join(join, one, operator, two, "RIGHT JOIN");
    }
    /**
     * 设置全局where条件
     * @param column
     * @param {StringOrNumber} operator
     * @param {StringOrNumber} value
     * @returns {Mysql}
     */
    where(column, operator, value) {
        if (column.constructor === Object) {
            for (let i in column) {
                this.where(i, '=', column[i]);
            }
            return this;
        }
        if (arguments.length === 2) {
            [value, operator] = [operator, '='];
        }
        column = mysql_1.escapeId(column);
        this.condition.where.push({ column, operator, value });
        return this;
    }
    /**
     * 设置or where条件
     * @param column
     * @param {StringOrNumber} operator
     * @param {StringOrNumber} value
     * @returns {Mysql}
     */
    orWhere(column, operator, value) {
        if (column.constructor === Object) {
            for (let i in column) {
                this.orWhere(i, '=', column[i]);
            }
            return this;
        }
        if (arguments.length === 2) {
            [value, operator] = [operator, '='];
        }
        column = mysql_1.escapeId(column);
        this.condition.orWhere.push({ column, operator, value });
        return this;
    }
    /**
     * 设置where null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    whereNull(colunm) {
        this.condition.whereNull.push(mysql_1.escapeId(colunm));
        return this;
    }
    /**
     * 设置where not null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    whereNotNull(colunm) {
        this.condition.whereNotNull.push(mysql_1.escapeId(colunm));
        return this;
    }
    /**
     * 设置or where null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    orWhereNull(colunm) {
        this.condition.orWhereNull.push(mysql_1.escapeId(colunm));
        return this;
    }
    /**
     * 设置or where not null字段
     * @param {string} colunm
     * @returns {Mysql}
     */
    orWhereNotNull(colunm) {
        this.condition.orWhereNotNull.push(mysql_1.escapeId(colunm));
        return this;
    }
    /**
     * 设置order by字段
     * @param {string} column
     * @param {"DESC" | "ASC"} direction
     * @returns {Mysql}
     */
    orderBy(column, direction = 'ASC') {
        this.condition.order.push(this.escapeIdSplit(column, ",") + ' ' + direction);
        return this;
    }
    /**
     * 设置跳过行数
     * @param {number} value
     * @returns {Mysql}
     */
    skip(value) {
        this.condition.skip = value;
        return this;
    }
    /**
     * 设置获取行数
     * @param {number} value
     * @returns {Mysql}
     */
    take(value) {
        this.condition.take = value;
        return this;
    }
    /**
     * 快捷设置分页参数
     * @param {number} skip
     * @param {number} take
     * @returns {Mysql}
     */
    limit(skip, take) {
        this.condition.skip = skip;
        if (take !== undefined) {
            this.condition.take = take;
        }
        return this;
    }
    /**
     * 设置GroupBy分组条件
     * @param {string} value
     * @returns {Mysql}
     */
    groupBy(value) {
        this.condition.groupBy = mysql_1.escapeId(value);
        return this;
    }
    /**
     * 查询多条数据
     * @returns {Mysql}
     */
    get() {
        let sql = this.toSelectSql();
        this.condition.sql = sql;
        return this;
    }
    /**
     * 查询总数（聚合）
     * @param {string} field
     * @returns {Mysql}
     */
    count(field) {
        this.condition.select = [this.escapeIdSplit(field, " as ", true)];
        this.get();
        return this;
    }
    /**
     * 查询单条数据
     * @returns {Mysql}
     */
    first() {
        this.skip(1).get();
        return this;
    }
    /**
     * 将Query转换为SQL
     * @returns {string}
     */
    toSql() {
        return this.clear();
    }
    /**
     * 将select转换为SQL
     * @returns {string}
     */
    toSelectSql() {
        let sql = '';
        if (this.condition.select.length < 1) {
            this.select('*');
        }
        sql += 'SELECT ' + this.condition.select.join(', ');
        sql += ' FROM ' + this.condition.table;
        sql += this.toJoin();
        sql += this.toWhere();
        sql += this.toOrder();
        if (this.condition.groupBy != undefined) {
            sql += ' GROUP BY ' + this.condition.groupBy;
        }
        if (this.condition.skip != undefined) {
            sql += ' LIMIT ' + this.condition.skip;
            if (this.condition.take != undefined) {
                sql += ', ' + this.condition.take;
            }
        }
        return sql;
    }
    /**
     * 多where条件组装
     * @param {Array<T>} whereMap
     * @param {(chunk: T) => string} handle
     * @returns {string}
     */
    whereHandle(whereMap, handle) {
        let whereList = [];
        whereList = whereMap.map((chunk) => handle(chunk));
        return whereList.join(' AND ');
    }
    /**
     * 组装where条件
     * @param {string} where
     * @param {string} condition
     * @returns {string}
     */
    joinWhere(where, condition) {
        where += (where === '' ? ' WHERE ' : ' AND ') + condition;
        return where;
    }
    /**
     * 组装orWhere条件
     * @param {string} where
     * @param {string} condition
     * @returns {string}
     */
    joinOrWhere(where, condition) {
        where += (where === '' ? ' OR ' : ' AND ') + condition;
        return where;
    }
    /**
     * 将where条件转换为SQL
     * @returns {string}
     */
    toWhere() {
        let where = '';
        let orWhere = '';
        if (this.condition.where.length > 0) {
            where = this.joinWhere(where, this.whereHandle(this.condition.where, chunk => {
                return chunk.column + ' ' + chunk.operator + ' ' + mysql_1.escape(chunk.value);
            }));
        }
        if (this.condition.whereNull.length > 0) {
            where = this.joinWhere(where, this.whereHandle(this.condition.whereNull, chunk => {
                return chunk + ' IS NULL';
            }));
        }
        if (this.condition.whereNotNull.length > 0) {
            where = this.joinWhere(where, this.whereHandle(this.condition.whereNotNull, chunk => {
                return chunk + ' IS NOT NULL';
            }));
        }
        if (this.condition.orWhere.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle(this.condition.orWhere, chunk => {
                return chunk.column + ' ' + chunk.operator + ' ' + mysql_1.escape(chunk.value);
            }));
        }
        if (this.condition.orWhereNull.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle(this.condition.orWhereNull, chunk => {
                return chunk + ' IS NULL';
            }));
        }
        if (this.condition.orWhereNotNull.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle(this.condition.orWhereNotNull, chunk => {
                return chunk + ' IS NOT NULL';
            }));
        }
        if (this.condition.findInSet.length > 0) {
            where = this.joinWhere(where, this.whereHandle(this.condition.findInSet, chunk => {
                return 'FIND_IN_SET(' + chunk.findStr + ',' + chunk.strList + ')';
            }));
        }
        if (this.condition.orFindInSet.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle(this.condition.orFindInSet, chunk => {
                return 'FIND_IN_SET(' + chunk.findStr + ',' + chunk.strList + ')';
            }));
        }
        if (this.condition.like.length > 0) {
            where = this.joinWhere(where, this.whereHandle(this.condition.like, chunk => {
                return chunk.field + ' LIKE ' + chunk.likeStr;
            }));
        }
        if (this.condition.orLike.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle(this.condition.orLike, chunk => {
                return chunk.field + ' LIKE ' + chunk.likeStr;
            }));
        }
        if (this.condition.whereIn.length > 0) {
            where = this.joinWhere(where, this.whereHandle(this.condition.whereIn, chunk => {
                return chunk.field + ' IN (' + chunk.whereIn.join(',') + ')';
            }));
        }
        if (this.condition.orWhereIn.length > 0) {
            orWhere = this.joinOrWhere(orWhere, this.whereHandle(this.condition.orWhereIn, chunk => {
                return chunk.field + ' IN (' + chunk.whereIn.join(',') + ')';
            }));
        }
        return where + orWhere;
    }
    /**
     * 将OrderBy条件转换为SQL
     * @returns {string}
     */
    toOrder() {
        let order = '';
        if (this.condition.order.length > 0) {
            order = ' ORDER BY ' + this.condition.order.join(', ');
        }
        return order;
    }
    /**
     * 将join转换为SQL
     * @returns {string}
     */
    toJoin() {
        let join = '';
        if (this.condition.join.length > 0) {
            let joinList = [];
            this.condition.join.forEach(chunk => {
                let joinListStr = ' ' + chunk.action + ' ' + chunk.join + ' ON ' + chunk.one + ' ' + chunk.operator + ' ' + chunk.two;
                if (this.condition.joinOnWhere.has(chunk.join)) {
                    let joinOnWhere = this.condition.joinOnWhere.get(chunk.join);
                    if (joinOnWhere.length > 0) {
                        let joinOnWhereList = [];
                        joinOnWhere.forEach(value => joinOnWhereList.push(value.column + ' ' + value.operator + ' ' + value.value));
                        joinListStr += ' AND ' + joinOnWhereList.join(' AND ');
                    }
                }
                joinList.push(joinListStr);
            });
            join += joinList.join(' ');
        }
        return join;
    }
    /**
     * 更新数据方法
     * @param {MysqlKV} column
     * @returns {Mysql}
     */
    update(column) {
        if (column.constructor !== Object) {
            throw new Error('parameter must be a Object');
        }
        /**
         * 使用对象拷贝
         * @type {MysqlKV}
         */
        const column_copy = Object.assign({}, column);
        let sql = 'UPDATE ' + this.condition.table;
        let update = [];
        for (let i in column_copy) {
            update.push(mysql_1.escapeId(i) + ' = ' + mysql_1.escape(column_copy[i]));
        }
        sql += ' SET ' + update.join(', ');
        sql += this.toWhere();
        this.condition.sql = sql;
        return this;
    }
    /**
     * 递增方法，可设置递增数量，默认为1
     * @param {string} field
     * @param {number} incrementNum
     * @returns {Mysql}
     */
    increment(field, incrementNum = 1) {
        let sql = '';
        sql += 'UPDATE ' + this.condition.table;
        field = mysql_1.escapeId(field);
        sql += ' SET ' + field + ' = ' + field + " + " + mysql_1.escape(incrementNum) + " ";
        sql += this.toWhere();
        this.condition.sql = sql;
        return this;
    }
    /**
     * 递减方法，可设置递减数量，默认为1
     * @param {string} field
     * @param {number} decrementNum
     * @returns {Mysql}
     */
    decrement(field, decrementNum = 1) {
        let sql = '';
        sql += 'UPDATE ' + this.condition.table;
        field = mysql_1.escapeId(field);
        sql += ' SET ' + field + ' = ' + field + " - " + mysql_1.escape(decrementNum) + " ";
        sql += this.toWhere();
        this.condition.sql = sql;
        return this;
    }
    /**
     * SQL中findInSet方法封装
     * @param {string | number} findStr
     * @param {string} strList
     * @returns {Mysql}
     */
    findInSet(findStr, strList) {
        this.condition.findInSet.push({
            findStr: mysql_1.escape(findStr),
            strList: mysql_1.escapeId(strList)
        });
        return this;
    }
    /**
     * SQL中or findInSet方法封装
     * @param {string | number} findStr
     * @param {string} strList
     * @returns {Mysql}
     */
    orFindInSet(findStr, strList) {
        this.condition.orFindInSet.push({
            findStr: mysql_1.escape(findStr),
            strList: mysql_1.escapeId(strList)
        });
        return this;
    }
    /**
     * like字符串匹配
     * @param {string} field
     * @param {string} likeStr
     * @returns {Mysql}
     */
    like(field, likeStr) {
        this.condition.like.push({
            field: mysql_1.escapeId(field),
            likeStr: mysql_1.escape(likeStr)
        });
        return this;
    }
    /**
     * or like字符串匹配
     * @param {string} field
     * @param {string} likeStr
     * @returns {Mysql}
     */
    orLike(field, likeStr) {
        this.condition.orLike.push({
            field: mysql_1.escapeId(field),
            likeStr: mysql_1.escape(likeStr)
        });
        return this;
    }
    /**
     * where in查询方法
     * @param {string} field
     * @param {Array<string | number>} whereIn
     * @returns {Mysql}
     */
    whereIn(field, whereIn) {
        /**
         * 使用数组拷贝
         * @type {string | number[]}
         */
        const whereIn_copy = [...whereIn];
        whereIn_copy.forEach((value, key) => whereIn_copy[key] = mysql_1.escape(value));
        this.condition.whereIn.push({
            field: mysql_1.escapeId(field),
            whereIn: whereIn_copy
        });
        return this;
    }
    /**
     * or where in查询方法
     * @param {string} field
     * @param {Array<string | number>} whereIn
     * @returns {Mysql}
     */
    orWhereIn(field, whereIn) {
        /**
         * 使用数组拷贝
         * @type {string | number[]}
         */
        const whereIn_copy = [...whereIn];
        whereIn_copy.forEach((value, key) => whereIn[key] = mysql_1.escape(value));
        this.condition.orWhereIn.push({
            field: mysql_1.escapeId(field),
            whereIn: whereIn_copy
        });
        return this;
    }
    /**
     * 删除数据方法
     * @returns {Mysql}
     */
    delete() {
        let sql = '';
        sql += 'DELETE FROM ' + this.condition.table;
        sql += this.toWhere();
        this.condition.sql = sql;
        return this;
    }
    /**
     * 插入数据方法
     * @param {MysqlKV} column
     * @returns {Mysql}
     */
    insert(column) {
        if (column.constructor !== Object) {
            throw new Error('parameter must be a Object');
        }
        /**
         * 使用对象拷贝
         * @type {MysqlKV}
         */
        const column_copy = Object.assign({}, column);
        let key = [];
        let value = [];
        for (let i in column_copy) {
            key.push(mysql_1.escapeId(i));
            value.push(mysql_1.escape(column_copy[i]));
        }
        let sql = 'INSERT INTO ' + this.condition.table;
        sql += '(' + key.join(', ') + ')';
        sql += ' VALUES(' + value.join(', ') + ')';
        this.condition.sql = sql;
        return this;
    }
    /**
     * 清空Query对象方法
     * @returns {string}
     */
    clear() {
        let sql = this.condition.sql;
        this.condition.clear();
        return sql;
    }
}
exports.Mysql = Mysql;
